@@ -23,17 +23,55 @@ struct sockaddr;
 namespace fz {
 class thread_pool;
 
+/** \brief The type of a socket event
+ *
+ * In received events, exactly a single bit is always set.
+ *
+ * Flag combinations are used when changing event handlers,
+ * \sa f::socket::set_event_handler
+ */
 enum class socket_event_flag
 {
-	// This is a nonfatal condition. It
-	// means there are additional addresses to try.
-	connection_next,
+	/**
+	 * Sent if connection attempt has failed, but there are still
+	 * additional addresses to try.
+	 * Errors can be ignored for this type.
+	 */
+	connection_next = 0x1,
 
-	connection,
-	read,
-	write
+	/**
+	 * If no error is set, the connection has been established.
+	 * If an error is set, the connection could not be established.
+	 */
+	connection = 0x2,
+
+	/**
+	 * If no error is set, data has become available.
+	 * If an error is set, the connection has failed.
+	 */
+	read = 0x4,
+
+	/**
+	 * If no error is set, data can be written.
+	 * If an error is set, the connection has failed.
+	 */
+	write = 0x8,
 };
 
+inline bool operator&(socket_event_flag lhs, socket_event_flag rhs) {
+	return (static_cast<std::underlying_type_t<socket_event_flag>>(lhs) & static_cast<std::underlying_type_t<socket_event_flag>>(rhs)) != 0;
+}
+inline socket_event_flag operator|(socket_event_flag lhs, socket_event_flag rhs)
+{
+	return static_cast<socket_event_flag>(static_cast<std::underlying_type_t<socket_event_flag>>(lhs) | static_cast<std::underlying_type_t<socket_event_flag>>(rhs));
+}
+inline socket_event_flag& operator|=(socket_event_flag& lhs, socket_event_flag rhs)
+{
+	lhs = lhs | rhs;
+	return lhs;
+}
+
+
 /**
  * \brief All classes sending socket events should derive from this.
  *
@@ -56,7 +94,7 @@ public:
 	}
 
 protected:
-	socket_event_source() = default;
+	socket_event_source() = delete;
 	explicit socket_event_source(socket_event_source* root)
 		: root_(root)
 	{}
@@ -74,13 +112,20 @@ struct socket_event_type;
  *
  * If the error value is non-zero for the connection, read and write events,
  * the socket has failed and needs to be closed. Doing anything else with
- * failed sockets is undefined behavior. 
+ * failed sockets is undefined behavior.
  * Failure events can be received at any time.
  *
  * Read and write events are edge-triggered:
- * - After receiving a read event for a socket, it will not be sent again unless
- * a subsequent call to socket::read has returned EAGAIN.
- * - The same holds for the write event and socket::write
+ * - After receiving a read event for a socket, it will not be sent again
+ *   unless a subsequent call to socket_interface::read or
+ *   socket_interface::shutdown_read has returned EAGAIN.
+ * - The same holds for the write event and socket_interface::write and
+ *   socket_interface::shutdown
+ * - A successful connection events doubles as write event, it does not
+ *   act as read event
+ *
+ * It is a grave violation to call the read/write/shutdown functions
+ * again after they returned EAGAIN without first waiting for the event.
  */
 typedef simple_event<socket_event_type, socket_event_source*, socket_event_flag, int> socket_event;
 
@@ -107,10 +152,11 @@ void FZ_PUBLIC_SYMBOL remove_socket_events(event_handler * handler, socket_event
  *
  * This function is called by socket::set_event_handler().
  *
- * \example Possible use-cases: Handoff after proxy handshakes, or handoff to TLS classes in
-			case of STARTTLS mechanism
+ * Example use-cases: Handoff after proxy handshakes, or handoff to TLS classes in case of STARTTLS mechanism
+ *
+ * Returns which events are still pending.
  */
-void FZ_PUBLIC_SYMBOL change_socket_event_handler(event_handler * old_handler, event_handler * new_handler, socket_event_source const* const source);
+fz::socket_event_flag FZ_PUBLIC_SYMBOL change_socket_event_handler(event_handler * old_handler, event_handler * new_handler, socket_event_source const* const source, fz::socket_event_flag remove);
 
 /// \private
 class socket_thread;
@@ -169,25 +215,23 @@ protected:
 
 	int close();
 
-	void do_set_event_handler(event_handler* pEvtHandler);
-
 	// Note: Unlocks the lock.
 	void detach_thread(scoped_lock & l);
 
 	thread_pool & thread_pool_;
 	event_handler* evt_handler_;
 
-	socket_t fd_{-1};
-
 	socket_thread* socket_thread_{};
 
+	socket_event_source * const ev_source_{};
+
+	socket_t fd_{-1};
+
 	unsigned int port_{};
 
 	int family_;
 
 	int buffer_sizes_[2];
-
-	socket_event_source * const ev_source_{};
 };
 
 class socket;
@@ -201,11 +245,40 @@ enum class listen_socket_state
 	listening,
 };
 
+/// Lightweight holder for socket descriptors
+class FZ_PUBLIC_SYMBOL socket_descriptor final
+{
+public:
+	socket_descriptor() = default;
+	~socket_descriptor();
+	explicit socket_descriptor(socket_base::socket_t fd) noexcept : fd_(fd) {}
+
+	socket_descriptor(socket_descriptor const&) = delete;
+	socket_descriptor& operator=(socket_descriptor const&) = delete;
+
+	socket_descriptor(socket_descriptor && rhs) noexcept { std::swap(fd_, rhs.fd_); }
+	socket_descriptor& operator=(socket_descriptor && rhs) noexcept {
+		std::swap(fd_, rhs.fd_);
+		return *this;
+	}
+
+	socket_base::socket_t detach() {
+		socket_base::socket_t ret = fd_;
+		fd_ = -1;
+		return ret;
+	}
+
+	explicit operator bool() const { return fd_ != -1; }
+
+private:
+	socket_base::socket_t fd_{-1};
+};
+
 /**
- * /brief Simple Listen socket
+ * \brief Simple Listen socket
  *
  * When a client connects, a socket event with the connection flag is send.
- * 
+ *
  * Call accept to accept.
  */
 class FZ_PUBLIC_SYMBOL listen_socket final : public socket_base, public socket_event_source
@@ -222,7 +295,7 @@ public:
 	/**
 	 * \brief Starts listening
 	 *
-	 * If no port is given, let the operating system devcide on a port. Can use local_port to query it afterwards.
+	 * If no port is given, let the operating system decide on a port. Can use local_port to query it afterwards.
 	 *
 	 * The address type, if not fz::address_type::unknown, must may the type of the address passed to bind()
 	 */
@@ -230,13 +303,19 @@ public:
 	int listen(address_type family, int port = 0);
 
 	/// Accepts incoming connection. If no socket is returned, error contains the reason
-	std::unique_ptr<socket> accept(int& error);
+	std::unique_ptr<socket> accept(int& error, fz::event_handler * handler = nullptr);
+
+	/**
+	 * \brief  Like accept, but only returns a socket descriptor.
+	 *
+	 * Best suited for tight accept loops where the descriptor is handed
+	 * off to other threads.
+	 */
+	socket_descriptor fast_accept(int& error);
 
 	listen_socket_state get_state() const;
 
-	void set_event_handler(event_handler* pEvtHandler) {
-		do_set_event_handler(pEvtHandler);
-	}
+	void set_event_handler(event_handler* pEvtHandler);
 
 private:
 	listen_socket_state state_{};
@@ -244,7 +323,7 @@ private:
 
 
 /// State transitions are monotonically increasing
-enum class socket_state
+enum class socket_state : unsigned char
 {
 	/// How the socket is initially
 	none,
@@ -274,7 +353,7 @@ enum class socket_state
 /**
  * \brief Interface for sockets
  *
- * Can be used for layers, see fz::socket for the expexted semantics.
+ * Can be used for layers, see fz::socket for the expected semantics.
  */
 class FZ_PUBLIC_SYMBOL socket_interface : public socket_event_source
 {
@@ -285,7 +364,45 @@ public:
 	virtual int read(void* buffer, unsigned int size, int& error) = 0;
 	virtual int write(void const* buffer, unsigned int size, int& error) = 0;
 
-	virtual void set_event_handler(event_handler* pEvtHandler) = 0;
+	template<typename T, std::enable_if_t<std::is_signed_v<T>, int> = 0>
+	int read(void* buffer, T size, int& error)
+	{
+		if (size < 0) {
+			error = EINVAL;
+			return -1;
+		}
+
+		return read(buffer, static_cast<unsigned int>(size), error);
+	}
+	template<typename T, std::enable_if_t<std::is_unsigned_v<T> && (sizeof(T) > sizeof(unsigned int)), int> = 0>
+	int read(void* buffer, T size, int& error)
+	{
+		if (size > std::numeric_limits<unsigned int>::max()) {
+			size = std::numeric_limits<unsigned int>::max();
+		}
+		return read(buffer, static_cast<unsigned int>(size), error);
+	}
+
+	template<typename T, std::enable_if_t<std::is_signed_v<T>, int> = 0>
+	int write(void const* buffer, T size, int& error)
+	{
+		if (size < 0) {
+			error = EINVAL;
+			return -1;
+		}
+
+		return write(buffer, static_cast<std::make_unsigned_t<T>>(size), error);
+	}
+	template<typename T, std::enable_if_t<std::is_unsigned_v<T> && (sizeof(T) > sizeof(unsigned int)), int> = 0>
+	int write(void const* buffer, T size, int& error)
+	{
+		if (size > std::numeric_limits<unsigned int>::max()) {
+			size = std::numeric_limits<unsigned int>::max();
+		}
+		return write(buffer, static_cast<unsigned int>(size), error);
+	}
+	
+	virtual void set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block = fz::socket_event_flag{}) = 0;
 
 	virtual native_string peer_host() const = 0;
 	virtual int peer_port(int& error) const = 0;
@@ -299,9 +416,14 @@ public:
 	 *
 	 * Only disallows further sends, does not affect reading from the
 	 * socket.
+	 *
+	 * Returns 0 on success, an error code otherwise.
+	 * If it returns EGAIN, shutdown is not yet complete. Call shutdown
+	 * again after the next write event.
 	 */
 	virtual int shutdown() = 0;
 
+	/// \see socket_layer::shutdown_read
 	virtual int shutdown_read() = 0;
 
 protected:
@@ -330,6 +452,8 @@ public:
 	socket(socket const&) = delete;
 	socket& operator=(socket const&) = delete;
 
+	static std::unique_ptr<socket> from_descriptor(socket_descriptor && desc, thread_pool & pool, int & error, fz::event_handler * handler = nullptr);
+
 	socket_state get_state() const override;
 	bool is_connected() const {
 		socket_state s = get_state();
@@ -410,15 +534,21 @@ public:
 	 */
 	int ideal_send_buffer_size();
 
-	/**
-	 * Allows re-triggering the read and write events.
-	 * Slow and cumbersome, use sparingly.
-	 */
-	void retrigger(socket_event_flag event);
-
 	virtual int shutdown() override;
 
-	virtual void set_event_handler(event_handler* pEvtHandler) override;
+	/**
+	 * \brief Changes the associated event handler.
+	 *
+	 * Pending events are rewritten to the new handler, or deleted if there is no new handler.
+	 *
+	 * Initially, the new handler is assumed to be waiting on read and write events and if the
+	 * socket is in a readable/writable state, the corresponding events are sent if not already
+	 * pending.
+	 *
+	 * As exception, events passed in retrigger_block are always removed and not resent if the socket
+	 * is in the readable/writable state.
+	 */
+	virtual void set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block = fz::socket_event_flag{}) override;
 
 	enum
 	{
@@ -430,6 +560,11 @@ public:
 	};
 
 	int flags() const { return flags_; }
+
+	/// Enables or disabled the passed flags
+	void set_flags(int flags, bool enable);
+
+	/// Sets the entire mask of enabled flag, disabling all others
 	void set_flags(int flags);
 
 	/**
@@ -446,10 +581,10 @@ private:
 	friend class listen_socket;
 	native_string host_;
 
-	socket_state state_{};
+	duration keepalive_interval_;
 
 	int flags_{};
-	duration keepalive_interval_;
+	socket_state state_{};
 };
 
 /**
@@ -464,6 +599,8 @@ private:
  *
  * For safe closing of a layer hierarchy, both the write and read side
  * should be shut down first, otherwise pending data might get discarded.
+ * The shutdown and shutdown_read functions may return EAGAIN, in which
+ * case they must be called again after the next write/read event.
  */
 class FZ_PUBLIC_SYMBOL socket_layer : public socket_interface
 {
@@ -475,19 +612,19 @@ public:
 	socket_layer& operator=(socket_layer const&) = delete;
 
 	/// The handler for any events generated (or forwarded) by this layer.
-	virtual void set_event_handler(event_handler* handler) override;
+	virtual void set_event_handler(event_handler* handler, fz::socket_event_flag retrigger_block = fz::socket_event_flag{}) override;
 
 	/**
-	 * Can be overriden to return something different, e.g. a proxy layer
+	 * Can be overridden to return something different, e.g. a proxy layer
 	 * would return the hostname of the peer connected to through the proxy,
-	 * whereas the next layer would return the address of the proxy istself.
+	 * whereas the next layer would return the address of the proxy itself.
 	 */
 	virtual native_string peer_host() const override { return next_layer_.peer_host(); }
 
 	/**
-	 * Can be overriden to return something different, e.g. a proxy layer
+	 * Can be overridden to return something different, e.g. a proxy layer
 	 * would return the post of the peer connected to through the proxy,
-	 * whereas the next layer would return the port of the proxy istself.
+	 * whereas the next layer would return the port of the proxy itself.
 	 */
 	virtual int peer_port(int& error) const override { return next_layer_.peer_port(error); }
 
@@ -502,7 +639,7 @@ public:
 	 *
 	 * \return 0 on success
 	 * \return EAGAIN if the shutdown cannot be completed, wait for a read event and try again.
-	 * \return otherwise an error has occured.
+	 * \return otherwise an error has occurred.
 	 *
 	 * On an ordinary socket, this is a no-op. Some layers however may return an EOF before the
 	 * next lower layer has reached its own EOF, such as the EOF of the secure channel from
@@ -517,6 +654,18 @@ public:
 	 */
 	virtual int shutdown_read() override;
 
+	virtual int connect(native_string const& host, unsigned int port, address_type family = address_type::unknown) override {
+		return next_layer_.connect(host, port, family);
+	}
+
+	virtual int shutdown() override {
+		return next_layer_.shutdown();
+	}
+
+	virtual socket_state get_state() const override {
+		return next_layer_.get_state();
+	}
+
 protected:
 	/**
 	 * Call in a derived classes handler for fz::socket_event. Results in
@@ -536,7 +685,7 @@ protected:
 	 * A pass-through layer does not handle events itself. Instead any events sent by the next layer get
 	 * sent to the event handler passed to the layer.
 	 */
-	void set_event_passthrough();
+	void set_event_passthrough(socket_event_flag retrigger_block = socket_event_flag{});
 
 	event_handler* event_handler_{};
 	socket_interface& next_layer_;
@@ -546,7 +695,7 @@ protected:
 /**
  * \brief Gets a symbolic name for socket errors.
  *
- * \example error_string(EAGAIN) == "EAGAIN"
+ * For example, <tt>error_string(EAGAIN) == "EAGAIN"</tt>
  *
  * \return name if the error code is known
  * \return number as string if the error code is not known
