@@ -16,6 +16,11 @@ class tls_layer;
 class tls_layer_impl;
 
 struct certificate_verification_event_type;
+
+/** \brief This event gets sent during the handshake with details about the session and the used certificate
+ *
+ * After receiving this event, \ref tls_layer::set_verification_result needs to be called eventually.
+ */
 typedef simple_event<certificate_verification_event_type, tls_layer*, tls_session_info> certificate_verification_event;
 
 /**
@@ -25,6 +30,10 @@ typedef simple_event<certificate_verification_event_type, tls_layer*, tls_sessio
  *
  * This class also supports TLS session resumption. Resumption has to be requested
  * explicitly, there is no shared state between unrelated sessions.
+ *
+ * Two trust models are possible with this class for client-side TLS: Certificates
+ * can either be evaluated against the system trust store, or you can implement a
+ * custom trust model such as TOFU.
  */
 class FZ_PUBLIC_SYMBOL tls_layer final : protected event_handler, public socket_layer
 {
@@ -39,8 +48,8 @@ public:
 	 *
 	 * If the handshake is started, wait for a connection event for the result.
 	 *
-	 * The certificate negotiated that eventually gets negotiated for the session]
-	 * must match the passed required_certificate, either in DER or PEM, 
+	 * The certificate that eventually gets negotiated for the session
+	 * must match the passed \c required_certificate, either in DER or PEM,
 	 * or the handshake will fail.
 	 */
 	bool client_handshake(std::vector<uint8_t> const& required_certificate, std::vector<uint8_t> const& session_to_resume = std::vector<uint8_t>(), native_string const& session_hostname = native_string());
@@ -52,11 +61,14 @@ public:
 	 *
 	 * If the handshake is started, wait for a connection event for the result.
 	 *
-	 * If a verification handler is passed, it will receive a 
-	 * certificate_verification_event event upon which it must call
-	 * set_verification_result.
-	 * If no verification handler is passed, verification is done soley using the system
+	 * If no verification handler is passed, verification is done solely using the system
 	 * trust store.
+	 *
+	 * If a verification handler is passed, it will receive a
+	 * \ref certificate_verification_event event upon which the handshake is paused until
+	 * \ref set_verification_result gets called including a \ref tls_session_info structure.
+	 * The handler is called even for certificates not trusted by the system trust store, allowing
+	 * the following impairments: Unknown issuer, wrong hostname and certificates used outside their validity time.
 	 */
 	bool client_handshake(event_handler *const verification_handler, std::vector<uint8_t> const& session_to_resume = std::vector<uint8_t>(), native_string const& session_hostname = native_string());
 
@@ -69,20 +81,13 @@ public:
 	 *
 	 * Before calling server_handshake, a valid certificate and key must be passed
 	 * in through \ref set_certificate.
+	 *
+	 * Session parameters of an existing session can be passed to allow session resumption. Check after handshake completion
+	 * with resumed_session()
+	 *
+	 * The preamble is sent out after setting up all the parameters, but before the first handshake message
 	 */
-	bool server_handshake();
-
-	/**
-	 * \brief Starts shaking hand for a new TLS session as server.
-	 *
-	 * Returns true if the handshake has started, false on error.
-	 *
-	 * If the handshake is started, wait for a connection event for the result.
-	 *
-	 * Before calling server_handshake, a valid certificate and key must be passed
-	 * in through \ref set_certificate.
-	 */
-	bool server_handshake(std::vector<uint8_t> const& session_to_resume);
+	bool server_handshake(std::vector<uint8_t> const& session_to_resume = {}, std::string_view const& preamble = {});
 
 	/// Gets session parameters for resumption
 	std::vector<uint8_t> get_session_parameters() const;
@@ -90,7 +95,11 @@ public:
 	/// Gets the session's peer certificate in DER
 	std::vector<uint8_t> get_raw_certificate() const;
 
-	/// Must be called after having received certificate_verification_event
+	/**
+	 * \brief Must be called after having received certificate_verification_event
+	 *
+	 * Can be used to trust a certificate even if it is not trusted via the system trust store.
+	 */
 	void set_verification_result(bool trusted);
 
 	std::string get_protocol() const;
@@ -124,8 +133,9 @@ public:
 	 *
 	 * If the pem flag is set, the input is assumed to be in PEM, otherwise DER.
 	 */
-	bool set_certificate(std::string const& key, std::string const& certs, native_string const& password, bool pem = true);
+	bool set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem = true);
 
+	/// Returns the version of the loaded GnuTLS library, may be different than the version used at compile-time.
 	static std::string get_gnutls_version();
 
 	/** \brief Creates a new private key and a self-signed certificate.
@@ -138,6 +148,21 @@ public:
 	 */
 	static std::pair<std::string, std::string> generate_selfsigned_certificate(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames);
 
+	/** \brief Negotiate application protocol
+	 *
+	 * If the peer makes use of ALPN, the handshake fails if no matching protocol is found.
+	 * If the peer does not use/support ALPN, the handshake continues and no protocol is negotiated.
+	 */
+	bool set_alpn(std::string_view const& alpn);
+	bool set_alpn(std::vector<std::string> const& alpns);
+
+	/// After a successful handshake, returns which protocol, if any, has been negotiated
+	std::string get_alpn() const;
+
+	/// If running as server, get the SNI sent by the client
+	native_string get_hostname() const;
+
+	bool is_server() const;
 
 	virtual socket_state get_state() const override;
 
@@ -150,6 +175,8 @@ public:
 
 	virtual int shutdown_read() override;
 
+	virtual void set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block = socket_event_flag{}) override;
+
 private:
 	virtual void FZ_PRIVATE_SYMBOL operator()(event_base const& ev) override;
 
